from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, List, NotRequired, Tuple, TypedDict

    class CoverageMeta(TypedDict):
        software: str
        version: str
        timestamp: str
        branch_coverage: bool
        show_contexts: bool

    class CoverageSummary(TypedDict):
        covered_lines: int
        missing_lines: int
        covered_branches: NotRequired[int]
        missing_branches: NotRequired[int]
        percent_covered: float

    class CoverageFile(TypedDict):
        executed_lines: List[int]
        missing_lines: List[int]
        executed_branches: NotRequired[List[Tuple[int, int]]]
        missing_branches: NotRequired[List[Tuple[int, int]]]
        summary: CoverageSummary

    class Coverage(TypedDict):
        meta: CoverageMeta
        files: Dict[str, CoverageFile]
        summary: CoverageSummary



def wrap_function(function):
    """Function wrapper to provide (monotonically increasing) coverage information
    while a test function is fuzzed"""

    import slipcover as sc

    sci = sc.Slipcover()
    # note no branch coverage support here (yet)
    sci.instrument(function)

    import functools

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        return function(sci, *args, **kwargs)

    return wrapper



import numpy as np
#import math

def testme():

    arr = [i for i in range(1,1000)]

    def doit1(x):
    #    x = [i*i for i in range(1,1000)][0]
        y = 1
    #    w, v = LA.eig(np.diag(arr)) # (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)))
        x = [i*i for i in range(0,100000)][99999]
        y1 = [i*i for i in range(0,200000)][199999]
        z1 = [i for i in range(0,300000)][299999]
        z = x * y
    #    z = np.multiply(x, y)
        return z

    def doit2(x):
        i = 0
    #    zarr = [math.cos(13) for i in range(1,100000)]
    #    z = zarr[0]
        z = 0.1
        while i < 100000:
    #        z = math.cos(13)
    #        z = np.multiply(x,x)
    #        z = np.multiply(z,z)
    #        z = np.multiply(z,z)
            z = z * z
            z = x * x
            z = z * z
            z = z * z
            i += 1
        return z

    def doit3(x):
        z = x + 1
        z = x + 1
        z = x + 1
        z = x + z
        z = x + z
    #    z = np.cos(x)
        return z

    def stuff():
        y = np.random.randint(1, 100, size=5000000)[4999999]
        x = 1.01
        for i in range(1,10):
            print(i)
            for j in range(1,10):
                x = doit1(x)
                x = doit2(x)
                x = doit3(x)
                x = 1.01
        return x

    stuff()

import sys
print("TESTME")
print(sys.argv)
testme()



def foo(n):
    r = 0
    if n >= 0:
        for i in range(1, n + 1):
            r += i
    return r


def test_some():
    assert 0 == foo(0)
    assert 1 == foo(1)
    assert 6 == foo(3)


def test_some_more():
    assert 10 == foo(4)



#!/usr/bin/env python3
"""Test the performance of pprint.PrettyPrinter.

This benchmark was available as `python -m pprint` until Python 3.12.

Authors: Fred Drake (original), Oleg Iarygin (pyperformance port).
"""

from time import perf_counter
from pprint import PrettyPrinter


printable = [('string', (1, 2), [3, 4], {5: 6, 7: 8})] * 100_000
p = PrettyPrinter()


if __name__ == '__main__':

    start_p = perf_counter()
    for i in range(7):
        p.pformat(printable)
    stop_p = perf_counter()
    print("Time elapsed: ", stop_p - start_p)



import sys  # just to see if it changes anything  # noqa: F401


class TestBase:
    def b(self, x):
        pass

    @classmethod
    def b_classm(cls, x):
        pass

    @staticmethod
    def b_static(x):
        pass


class Test(TestBase):
    def f1(self, x):
        pass

    class Inner:
        def f2(self, x):
            pass

        def f3(self, x):
            pass

        class InnerInner:
            def f4(self, x):
                pass

    @staticmethod
    def f_static(x):
        pass

    @classmethod
    def f_classm(cls, x):
        pass


def f5():
    def f6():
        pass


def f7():
    pass


if __name__ == "__main__":
    Test.Inner().f2(0)



import pytest
import ast
import covers.branch as br


def ast_parse(s):
    import inspect

    return ast.parse(inspect.cleandoc(s))


def get_branches(code):
    from covers.covers import Covers

    return sorted(Covers.branches_from_code(code))


def check_locations(node: ast.AST):
    """Implements the same checks as the VALIDATE_POSITIONS checks from Python/ast.c (Python 3.12.3)"""
    for attr in node._attributes:
        assert hasattr(node, attr)

    if "end_lineno" in node._attributes:
        assert node.end_lineno >= node.lineno
        if node.lineno < 0:
            assert node.end_lineno == node.lineno

    if "end_col_offset" in node._attributes:
        if node.col_offset < 0:
            assert node.end_col_offset == node.col_offset

    if "end_lineno" in node._attributes and "end_col_offset" in node._attributes:
        if node.lineno == node.end_lineno:
            assert node.end_col_offset >= node.col_offset

    for child in ast.iter_child_nodes(node):
        check_locations(child)


def test_if():
    source = """
        if x == 0:
            x += 2

        x += 3
    """
    import inspect

    source = inspect.cleandoc(source)
    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(1, 2), (1, 4)] == get_branches(code)


def test_if_else():
    import inspect

    source = """
        if x == 0:
            x += 1

        else:

            x += 2

        x += 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(1, 2), (1, 6)] == get_branches(code)


def test_if_elif_else():
    import inspect

    source = """
        if x == 0:
            x += 1
        elif x == 1:
            x += 2
        else:
            x += 3

        x += 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(1, 2), (1, 3), (3, 4), (3, 6)] == get_branches(code)


def test_if_nothing_after_it():
    import inspect

    source = """
        if x == 0:
            x += 1

    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(1, 0), (1, 2)] == get_branches(code)


def test_if_nested():
    import inspect

    source = """
        if x >= 0:
            y = 1
            if x > 1:
                if x > 2:
                    y = 2
        else:
            y = 4
            if x < 1:
                y = 3
            y += 10
        z = 0
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [
        (1, 2),
        (1, 7),
        (3, 4),
        (3, 11),
        (4, 5),
        (4, 11),
        (8, 9),
        (8, 10),
    ] == get_branches(code)


def test_if_in_function():
    import inspect

    source = """
        def foo(x):
            if x >= 0:
                return 1

        async def bar(x):
            if x == 0:
                return 1

        class Foo:
            def __init__(self, x):
                if x == 0:
                    self.x = 0

        foo(-1)
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 0), (2, 3), (6, 0), (6, 7), (11, 0), (11, 12)] == get_branches(code)


def test_keep_docstrings():
    import inspect

    source = """
        def foo(x):
            \"\"\"foo something\"\"\"
            if x >= 0:
                return 1

        async def bar(x):
            \"\"\"bar something\"\"\"
            if x == 0:
                return 1

        class Foo:
            \"\"\"Foo something\"\"\"
            def __init__(self, x):
                if x == 0:
                    self.x = 0

        foo(-1)
    """
    source = inspect.cleandoc(source)
    #    print(ast.dump(t, indent=True))

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(3, 0), (3, 4), (8, 0), (8, 9), (14, 0), (14, 15)] == get_branches(code)

    g = {}
    exec(code, g, g)

    assert "foo something" == g["foo"].__doc__
    assert "bar something" == g["bar"].__doc__
    assert "Foo something" == g["Foo"].__doc__


def test_branch_bad_positions():
    import inspect

    source = """\
def foo(x):
    if x == 0:
        t = '''\\
x == 0
'''
    else:
        t = '''\\
x != 0
'''
"""
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    # on Python 3.12.3, compile() raises a ValueError due to failing
    # the VALIDATE_POSITIONS checks from Python/ast.c
    compile(t, "foo.py", "exec")
    check_locations(t)


def test_for():
    import inspect

    source = """
        for v in [1, 2]:
            if v > 0:
                x += v

        x += 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(1, 2), (1, 5), (2, 1), (2, 3)] == get_branches(code)


def test_async_for():
    import inspect

    source = """
        import asyncio

        async def fun():
            global x

            async def g():
                yield 1
                yield 2

            async for v in g(): #10
                if v > 0:
                    x += v
            x += 3

        asyncio.run(fun())
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(10, 11), (10, 13), (11, 12), (11, 13)] == get_branches(code)


def test_for_else():
    import inspect

    source = """
        for v in [1, 2]:
            if v > 0:
                x += v
        else:
            x += 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(1, 2), (1, 5), (2, 1), (2, 3)] == get_branches(code)


def test_for_break_else():
    import inspect

    source = """
        for v in [1, 2]:
            if v > 0:
                x += v
            break
        else:
            x += 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(1, 2), (1, 6), (2, 3), (2, 4)] == get_branches(code)


def test_while():
    import inspect

    source = """
        v = 2
        while v > 0:
            v -= 1
            if v > 0:
                x += v

        x += 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 3), (2, 7), (4, 2), (4, 5)] == get_branches(code)


def test_while_else():
    import inspect

    source = """
        v = 2
        while v > 0:
            v -= 1
            if v > 0:
                x += v
        else:
            x += 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 3), (2, 7), (4, 2), (4, 5)] == get_branches(code)


def test_while_break_else():
    import inspect

    source = """
        v = 2
        while v > 0:
            v -= 1
            if v > 0:
                x += v
            break
        else:
            x += 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 3), (2, 8), (4, 5), (4, 6)] == get_branches(code)


def test_match():
    import inspect

    source = """
        v = 2
        match v:
            case 1:
                x = 1
            case 2:
                x = 2
        x += 2
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 4), (2, 6), (2, 7)] == get_branches(code)


def test_match_case_with_false_guard():
    import inspect

    source = """
        x = 0
        v = 1
        match v:
            case 1 if x > 0:
                x = 1
            case 1:
                x = 2
        x += 2
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(3, 5), (3, 7), (3, 8)] == get_branches(code)


def test_match_case_with_guard_isnt_wildcard():
    import inspect

    source = """
        def fun(v):
            match v:
                case _ if v > 0:
                    print("not default")
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 0), (2, 4)] == get_branches(code)


def test_match_branch_to_exit():
    import inspect

    source = """
        v = 5
        match v:
            case 1:
                x = 1
            case 2:
                x = 2
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 0), (2, 4), (2, 6)] == get_branches(code)


def test_match_default():
    import inspect

    source = """
        v = 5
        match v:
            case 1:
                x = 1
            case 2:
                x = 2
            case _:
                x = 3
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 4), (2, 6), (2, 8)] == get_branches(code)


def test_branch_after_case():
    import inspect

    source = """
        v = 1
        match v:
            case 1:
                if x < 0:  #4
                    x = 1
            case 2:
                if x < 0:  #7
                    x = 1
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 0), (2, 4), (2, 7), (4, 0), (4, 5), (7, 0), (7, 8)] == get_branches(
        code
    )


def test_branch_after_case_with_default():
    import inspect

    source = """
        v = 1
        match v:
            case 1:
                if x < 0:  #4
                    x = 1
            case 2:
                if x < 0:  #7
                    x = 1
            case _:
                if x < 0:  #10
                    x = 1
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [
        (2, 4),
        (2, 7),
        (2, 10),
        (4, 0),
        (4, 5),
        (7, 0),
        (7, 8),
        (10, 0),
        (10, 11),
    ] == get_branches(code)


def test_branch_after_case_with_next():
    import inspect

    source = """
        v = 1
        match v:
            case 1:
                if x < 0:  #4
                    x = 1
            case 2:
                if x < 0:  #7
                    x = 1
        x += 1
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 4), (2, 7), (2, 9), (4, 5), (4, 9), (7, 8), (7, 9)] == get_branches(
        code
    )


def test_match_wildcard_in_match_or():
    # Thanks to Ned Batchelder for this test case
    import inspect

    source = """
            def absurd(x):
                match x:
                    case (3 | 99 | (999 | _)):
                        print("default")
            absurd(5)
            """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 4)] == get_branches(code)


def test_match_capture():
    import inspect

    source = """
            def capture(x):
                match x:
                    case y:
                        print("default")
            capture(5)
            """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(2, 4)] == get_branches(code)


@pytest.mark.parametrize("star", ["", "*"])
def test_try_except(star):
    import inspect

    source = f"""
        def foo(x):
            try:
                y = x + 1
                if y < 0:
                    y = 0
            except{star} RuntimeException:
                if y < 2:
                    y = 0
            except{star} FileNotFoundError:
                if y < 2:
                    y = 0

            return 2*y
    """

    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(4, 5), (4, 13), (7, 8), (7, 13), (10, 11), (10, 13)] == get_branches(code)


def test_try_finally():
    import inspect

    source = """
        def foo(x):
            try:
                y = x + 1
                if y < 0:
                    y = 0
            finally:
                y = 2*y
    """
    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(4, 5), (4, 7)] == get_branches(code)


@pytest.mark.parametrize("star", ["", "*"])
def test_try_else(star):
    import inspect

    source = f"""
        def foo(x):
            try:
                y = x + 1
                if y < 0:
                    y = 0
            except{star} RuntimeException:
                if y < 2:
                    y = -1
            else:
                y = 2*y
    """

    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(4, 5), (4, 10), (7, 0), (7, 8)] == get_branches(code)


@pytest.mark.parametrize("star", ["", "*"])
def test_try_else_finally(star):
    import inspect

    source = f"""
        def foo(x):
            try:
                y = x + 1
                if y < 0:
                    y = 0
            except{star} RuntimeException:
                if y < 2:
                    y = -1
            else:
                if y > 5:
                    y = 42
            finally:
                y = 2*y
    """

    source = inspect.cleandoc(source)

    t = br.preinstrument(source)
    check_locations(t)
    code = compile(t, "foo", "exec")
    assert [(4, 5), (4, 10), (7, 8), (7, 13), (10, 11), (10, 13)] == get_branches(code)



# Compares the "missing" attribute of files in a coverage.py
# (or slipcover) JSON output.

import json
import sys

file = (sys.argv[1], sys.argv[2])
data = [None, None]

for i in [0, 1]:
    with open(file[i], "r") as f:
        data[i] = json.load(f)

def merge_consecutives(L):
    # Neat little trick due to John La Rooy: the difference between the numbers
    # on a list and a counter is constant for consecutive items :)
    from itertools import groupby, count

    groups = groupby(sorted(L), key=lambda item, c=count(): item - next(c))
    return [
        str(g[0]) if g[0] == g[-1] else f"{g[0]}-{g[-1]}"
        for g in [list(g) for _, g in groups]
    ]

for (f_name, f0_cover) in data[0]['files'].items():
    if f_name in data[1]['files']:
        for item in ('executed_lines', 'missing_lines'):
            f0_item = set(f0_cover[item])
            f1_item = set(data[1]['files'][f_name][item])
            if f0_item != f1_item:
                print(f"{f_name}  {item}  differs:")
                if f0_item - f1_item:
                    print(f"    only in {file[0]}: {merge_consecutives(f0_item - f1_item)}")
                if f1_item - f0_item:
                    print(f"    only in {file[1]}: {merge_consecutives(f1_item - f0_item)}")
                print("")

def is_pytest(filename):
    from pathlib import Path
    import re

    p = Path(filename)
    # pytest file conventions
    if p.name == 'conftest.py' or re.match(r"test_.*\.py$", p.name) or re.match(r".*_test\.py$", p.name):
           return True

#    with open(filename, "r") as f:
#        if re.search(r"^(import pytest|from pytest )", f.read(), re.M):
#            return True

    return False

for is_test in [False, True]:
    for i in [0, 1]:
        other = 1-i
        any = False

        for f_name in sorted(data[i]['files'].keys() - data[other]['files'].keys()):
            if is_pytest(f_name) != is_test:
                continue

            if not any:
                print(("tests " if is_test else "") + f"only in {file[i]}:")
                any = True

            print(f"    {f_name}")

        if any:
            print("")



from __future__ import annotations

import dis
import types
from typing import TYPE_CHECKING

# Import from Rust
from .covers_core import (  # noqa: F401
    Covers,
    CoverageTracker,
    PathSimplifier,
    add_summaries,
    branches_from_code,
    decode_branch,
    encode_branch,
    format_missing_py as format_missing,
    is_branch,
    lines_from_code,
    merge_coverage,
    print_coverage,
    print_xml,
    print_lcov,
    __version__,
    # Bytecode classes
    Branch,
    Editor,
    ExceptionTableEntry,
    LineEntry,
    # Exceptions
    CoversError,
)

__all__ = [
    # Core classes (from Rust)
    "Covers",
    "CoverageTracker",
    "PathSimplifier",
    # Branch functions (from Rust)
    "encode_branch",
    "decode_branch",
    "is_branch",
    # Code analysis functions (from Rust)
    "lines_from_code",
    "branches_from_code",
    # Reporting functions (from Rust)
    "add_summaries",
    "print_coverage",
    "print_xml",
    "print_lcov",
    # Bytecode classes (from Rust)
    "Branch",
    "Editor",
    "ExceptionTableEntry",
    "LineEntry",
    # Python utilities
    "findlinestarts",
    "format_missing",
    "merge_coverage",
    # Exceptions
    "CoversError",
    # Version
    "__version__",
]

# Python 3.13 returns 'None' lines;
# Python 3.11+ generates a line just for RESUME or RETURN_GENERATOR, POP_TOP, RESUME;
# Python 3.11 generates a 0th line
_op_RESUME = dis.opmap["RESUME"]
_op_RETURN_GENERATOR = dis.opmap["RETURN_GENERATOR"]


def findlinestarts(co: types.CodeType):
    for off, line in dis.findlinestarts(co):
        if line and co.co_code[off] not in (_op_RESUME, _op_RETURN_GENERATOR):
            yield off, line


if TYPE_CHECKING:
    pass


# format_missing is now implemented in Rust (imported above)
# print_xml is now implemented in Rust (imported above)
# print_coverage is now implemented in Rust (imported above)
# merge_coverage is now implemented in Rust (imported above)
# CoversError is now implemented in Rust (imported above)


# The Covers class is now implemented in Rust (covers_core)
# All methods are available from the imported class above



from pathlib import Path
import sys
import re

sys.path += str(Path(sys.argv[0]).parent)
from benchmarks import BENCHMARK_JSON, cases

def parse_args():
    import argparse
    ap = argparse.ArgumentParser()

    ap.add_argument('--os', type=str, default='Linux', help='select OS name')
    ap.add_argument('--case', choices=[c.name for c in cases],
                       action='extend', nargs='+', help='select case(s) to run/plot')
    ap.add_argument('--bench', action='extend', nargs='+', help='select benchmark(s) to plot')
    ap.add_argument('--title', type=str, default='Overhead by Python version', help='set plot title')
    ap.add_argument('--out', type=Path, default=Path("benchmarks/versions.png"), help='set plot output file')
    ap.add_argument('--latex', type=Path, help='also output LaTeX table to given file')
    ap.add_argument('--style', type=str, help='set matplotlib style')
    ap.add_argument('--figure-width', type=float, default=12, help='matplotlib figure width')
    ap.add_argument('--figure-height', type=float, default=8, help='matplotlib figure height')
    ap.add_argument('--bar-labels', action='store_true', help='add labels to bars')
    ap.add_argument('--font-size-delta', type=int, default=0, help='increase or decrease font size')
    ap.add_argument('--rename-slipcover', type=str, help='rename SlipCover in names to given string')
    ap.add_argument('--skip-version', default=[], action='append', help='omit given Python version')
    ap.add_argument('--absolute', action='store_true', help='emit absolute numbers in LaTeX')
    ap.add_argument('--yscale', type=str, default="linear", help='set matplotlib Y scale')

    ap.add_argument('--use-tex', type=str, help='Selects to use (La)TeX fonts and specifies the font family to use')

    args = ap.parse_args()

    if not args.case:
        args.case = ['coveragepy', 'coveragepy-branch', 'slipcover', 'slipcover-branch']

    if not args.bench:
        args.bench = ['raytrace']

    return args

args = parse_args()

def load_data():
    import json
    import packaging.version

    with open(BENCHMARK_JSON, 'r') as f:
        entries = json.load(f)

    entries = [e for e in entries if e['system']['os'][0].casefold() == args.os.casefold()]

    os_versions = set(e['system']['os'][1] for e in entries)
    if len(os_versions) > 1:
        # pick "latest" OS version... FIXME pick by benchmark date, as in benchmarks.py
        latest = max(os_versions)
        entries = [e for e in entries if e['system']['os'][1] == latest]

    systems = set([('/'.join(e['system']['os']), e['system']['cpu']) for e in entries])
    if len(systems) > 1:
        raise RuntimeException(f"Results sets mixes data from different systems: {systems}")

    # Python version -> results
    v2r = {e['system']['python']: e['results'] for e in entries}

    for v in args.skip_version:
        del v2r[v]

    # sort Python versions semantically
    python_versions = sorted(v2r.keys(), key=lambda x: packaging.version.parse(x))
    print("versions: ", python_versions)

    # check data looks ok -- all should have the same cases
    common_cases = set.intersection(*(set(r.keys()) for r in v2r.values()))
    print("common cases: ", common_cases)
    assert common_cases == set(v2r[python_versions[0]].keys())

    if 'all' in args.bench:
        common_benchmarks = set.intersection(*(set(b.keys()) for r in v2r.values() for b in r.values()))
        print("common benchmarks: ", common_benchmarks)
        args.bench.remove('all')
        args.bench.extend(sorted(common_benchmarks))

    # check that requested benchmark(s) are available everywhere
    for v, b in ((v, b) for v in v2r for b in v2r[v].values()):
        for bench in args.bench:
            if bench not in b:
                print(f"Benchmark {bench} not available for {v}")
                sys.exit(1)

    # compute the median
    from statistics import median
    for v in python_versions:
        for c in v2r[v]:
            for b in v2r[v][c]:
                m = v2r[v][c][b]
                m['median'] = median(m['times'])

    return v2r, python_versions

v2r, python_versions = load_data()


def latex_results(args):
    selected_cases = [c for c in cases if c.name in args.case]
    nonbase_cases = [c for c in selected_cases if c.name != 'base']

    assert len(args.bench) == 1, "Only one benchmark at a time supported."

    def latex_escape(s):
        repl = {
            '&':  r'\&',
            '%':  r'\%', 
            '$':  r'\$', 
            '#':  r'\#', 
            '_':  r'\_', 
            '{':  r'\{', 
            '}':  r'\}',
            '~':  r'\textasciitilde{}', 
            '^':  r'\^{}', 
            '\\': r'\textbackslash',
        }

        return "".join([repl.get(c, c) for c in s])

    def texttt(s):
        return f"\\texttt{{{s}}}"

    with open(args.latex, "w") as out:
        print("\\begin{tabular}{l " + ("r " * (int(args.absolute)+len(nonbase_cases))) + "}", file=out)
        line = "\\thead[l]{Python\\\\version}"

        if args.absolute:
            line += " & \\thead[r]{no coverage}"

        for case in nonbase_cases:
            case_name = re.sub('[Ss]lip[Cc]over', '\\\\systemname{}', latex_escape(case.label))
            case_name = re.sub('coverage\\.py', '\\\\texttt{coverage.py}', case_name)

            import textwrap
            case_name = "\\\\ ".join(textwrap.wrap(case_name, 10, break_long_words=False))

            line += " & \\thead[r]{" + case_name + "}"
        line += " \\\\"
        print(line, file=out)
        print("\\hline", file=out)

        for version in python_versions:
            line = f"{texttt(latex_escape(version))}"

            base_result = v2r[version]['base'][args.bench[0]]['median']

            if args.absolute:
                line += f" & \\SI{{{base_result:.1f}}}{{s}}"

            for case in nonbase_cases:
                if args.absolute:
                    r = v2r[version][case.name][args.bench[0]]['median']
                    line += f" & \\SI{{{r:.1f}}}{{s}}"
                else:
                    r = v2r[version][case.name][args.bench[0]]['median'] / base_result
                    line += f" & {r:.2f}$\\times$"
            line += " \\\\"
            print(line, file=out)

        print("\\hline", file=out)
        print("\\end{tabular}", file=out)

    print(f"Wrote to {args.latex}.")
    print("")


nonbase_cases = [c for c in cases if c.name != 'base' and c.name in args.case]

import matplotlib.pyplot as plt
import numpy as np

plt.rcParams.update({
    'font.weight': 'bold',
    'pdf.fonttype': 42  # output TrueType; bigger but scalable
})
if args.use_tex:
    plt.rcParams.update({
        "text.usetex": True,
        "font.family": args.use_tex
    })
plt.rc('ytick', labelsize=12+args.font_size_delta)

if args.style:
    plt.style.use(args.style)

x = np.arange(len(python_versions))

fig, ax = plt.subplots()

ax.set_yscale(args.yscale)
if args.yscale == 'log':
    from matplotlib.ticker import ScalarFormatter
    ax.yaxis.set_major_formatter(ScalarFormatter())

if len(args.bench) == 1: # do a bar plot
    n_bars = len(nonbase_cases)
    width = .70 # of all bars
    bars_x = np.arange(width/n_bars/2, width, width/n_bars) - width/2

    for case, bar_x in zip(nonbase_cases, bars_x):
        r = [v2r[v][case.name][args.bench[0]]['median'] / v2r[v]['base'][args.bench[0]]['median'] for v in python_versions]

        case_label = case.label
        if args.rename_slipcover:
            case_label = re.sub('[Ss]lip[Cc]over', args.rename_slipcover, case_label)

        rects = ax.bar(x + bar_x, r, width/n_bars, label=case_label, color=case.color, zorder=2)

        if args.bar_labels:
            ax.bar_label(rects, padding=3, labels=[f'{v:.1f}x' for v in r], fontsize=8+args.font_size_delta)

else: # do a box plot
    for bench, case in ((bench, case) for bench in args.bench for case in nonbase_cases):
        r = [v2r[v][case.name][bench]['median'] / v2r[v]['base'][bench]['median'] for v in python_versions]

        ax.plot(r, label=f"{bench} {case.name}")

ax.set_title(args.title, size=18+args.font_size_delta, weight='bold')
ax.set_ylabel('Normalized execution time', size=15+args.font_size_delta)
ax.set_xticks(x, labels=python_versions, fontsize=15+args.font_size_delta)
if not args.style:
    ax.grid(axis='y', alpha=.3)
ax.axhline(y=1, color='black', linewidth=1, alpha=.5, zorder=1)
ax.legend(fontsize=15+args.font_size_delta)

fig.set_size_inches(args.figure_width, args.figure_height)
fig.tight_layout()
fig.savefig(args.out)
print(f"Plotted to {args.out}.")
print("")

if args.latex:
    latex_results(args)



from array import array
import math

import pyperf
from six.moves import xrange


class Array2D(object):

    def __init__(self, w, h, data=None):
        self.width = w
        self.height = h
        self.data = array('d', [0]) * (w * h)
        if data is not None:
            self.setup(data)

    def _idx(self, x, y):
        if 0 <= x < self.width and 0 <= y < self.height:
            return y * self.width + x
        raise IndexError

    def __getitem__(self, x_y):
        (x, y) = x_y
        return self.data[self._idx(x, y)]

    def __setitem__(self, x_y, val):
        (x, y) = x_y
        self.data[self._idx(x, y)] = val

    def setup(self, data):
        for y in xrange(self.height):
            for x in xrange(self.width):
                self[x, y] = data[y][x]
        return self

    def indexes(self):
        for y in xrange(self.height):
            for x in xrange(self.width):
                yield x, y

    def copy_data_from(self, other):
        self.data[:] = other.data[:]


class Random(object):
    MDIG = 32
    ONE = 1
    m1 = (ONE << (MDIG - 2)) + ((ONE << (MDIG - 2)) - ONE)
    m2 = ONE << MDIG // 2
    dm1 = 1.0 / float(m1)

    def __init__(self, seed):
        self.initialize(seed)
        self.left = 0.0
        self.right = 1.0
        self.width = 1.0
        self.haveRange = False

    def initialize(self, seed):

        self.seed = seed
        seed = abs(seed)
        jseed = min(seed, self.m1)
        if (jseed % 2 == 0):
            jseed -= 1
        k0 = 9069 % self.m2
        k1 = 9069 / self.m2
        j0 = jseed % self.m2
        j1 = jseed / self.m2
        self.m = array('d', [0]) * 17
        for iloop in xrange(17):
            jseed = j0 * k0
            j1 = (jseed / self.m2 + j0 * k1 + j1 * k0) % (self.m2 / 2)
            j0 = jseed % self.m2
            self.m[iloop] = j0 + self.m2 * j1
        self.i = 4
        self.j = 16

    def nextDouble(self):
        I, J, m = self.i, self.j, self.m
        k = m[I] - m[J]
        if (k < 0):
            k += self.m1
        self.m[J] = k

        if (I == 0):
            I = 16
        else:
            I -= 1
        self.i = I

        if (J == 0):
            J = 16
        else:
            J -= 1
        self.j = J

        if (self.haveRange):
            return self.left + self.dm1 * float(k) * self.width
        else:
            return self.dm1 * float(k)

    def RandomMatrix(self, a):
        for x, y in a.indexes():
            a[x, y] = self.nextDouble()
        return a

    def RandomVector(self, n):
        return array('d', [self.nextDouble() for i in xrange(n)])


def copy_vector(vec):
    # Copy a vector created by Random.RandomVector()
    vec2 = array('d')
    vec2[:] = vec[:]
    return vec2


class ArrayList(Array2D):

    def __init__(self, w, h, data=None):
        self.width = w
        self.height = h
        self.data = [array('d', [0]) * w for y in xrange(h)]
        if data is not None:
            self.setup(data)

    def __getitem__(self, idx):
        if isinstance(idx, tuple):
            return self.data[idx[1]][idx[0]]
        else:
            return self.data[idx]

    def __setitem__(self, idx, val):
        if isinstance(idx, tuple):
            self.data[idx[1]][idx[0]] = val
        else:
            self.data[idx] = val

    def copy_data_from(self, other):
        for l1, l2 in zip(self.data, other.data):
            l1[:] = l2


def SOR_execute(omega, G, cycles, Array):
    for p in xrange(cycles):
        for y in xrange(1, G.height - 1):
            for x in xrange(1, G.width - 1):
                G[x, y] = (omega * 0.25 * (G[x, y - 1] + G[x, y + 1] + G[x - 1, y]
                                           + G[x + 1, y])
                           + (1.0 - omega) * G[x, y])


def bench_SOR(loops, n, cycles, Array):
    range_it = xrange(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        G = Array(n, n)
        SOR_execute(1.25, G, cycles, Array)

    return pyperf.perf_counter() - t0


def SparseCompRow_matmult(M, y, val, row, col, x, num_iterations):
    range_it = xrange(num_iterations)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        for r in xrange(M):
            sa = 0.0
            for i in xrange(row[r], row[r + 1]):
                sa += x[col[i]] * val[i]
            y[r] = sa

    return pyperf.perf_counter() - t0


def bench_SparseMatMult(cycles, N, nz):
    x = array('d', [0]) * N
    y = array('d', [0]) * N

    nr = nz // N
    anz = nr * N
    val = array('d', [0]) * anz
    col = array('i', [0]) * nz
    row = array('i', [0]) * (N + 1)

    row[0] = 0
    for r in xrange(N):
        rowr = row[r]
        step = r // nr
        row[r + 1] = rowr + nr
        if step < 1:
            step = 1
        for i in xrange(nr):
            col[rowr + i] = i * step

    return SparseCompRow_matmult(N, y, val, row, col, x, cycles)


def MonteCarlo(Num_samples):
    rnd = Random(113)
    under_curve = 0
    for count in xrange(Num_samples):
        x = rnd.nextDouble()
        y = rnd.nextDouble()
        if x * x + y * y <= 1.0:
            under_curve += 1
    return float(under_curve) / Num_samples * 4.0


def bench_MonteCarlo(loops, Num_samples):
    range_it = xrange(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        MonteCarlo(Num_samples)

    return pyperf.perf_counter() - t0


def LU_factor(A, pivot):
    M, N = A.height, A.width
    minMN = min(M, N)
    for j in xrange(minMN):
        jp = j
        t = abs(A[j][j])
        for i in xrange(j + 1, M):
            ab = abs(A[i][j])
            if ab > t:
                jp = i
                t = ab
        pivot[j] = jp

        if A[jp][j] == 0:
            raise Exception("factorization failed because of zero pivot")

        if jp != j:
            A[j], A[jp] = A[jp], A[j]

        if j < M - 1:
            recp = 1.0 / A[j][j]
            for k in xrange(j + 1, M):
                A[k][j] *= recp

        if j < minMN - 1:
            for ii in xrange(j + 1, M):
                for jj in xrange(j + 1, N):
                    A[ii][jj] -= A[ii][j] * A[j][jj]


def LU(lu, A, pivot):
    lu.copy_data_from(A)
    LU_factor(lu, pivot)


def bench_LU(cycles, N):
    rnd = Random(7)
    A = rnd.RandomMatrix(ArrayList(N, N))
    lu = ArrayList(N, N)
    pivot = array('i', [0]) * N
    range_it = xrange(cycles)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        LU(lu, A, pivot)

    return pyperf.perf_counter() - t0


def int_log2(n):
    k = 1
    log = 0
    while k < n:
        k *= 2
        log += 1
    if n != 1 << log:
        raise Exception("FFT: Data length is not a power of 2: %s" % n)
    return log


def FFT_num_flops(N):
    return (5.0 * N - 2) * int_log2(N) + 2 * (N + 1)


def FFT_transform_internal(N, data, direction):
    n = N // 2
    bit = 0
    dual = 1
    if n == 1:
        return

    logn = int_log2(n)
    if N == 0:
        return
    FFT_bitreverse(N, data)

    # apply fft recursion
    # this loop executed int_log2(N) times
    bit = 0
    while bit < logn:
        w_real = 1.0
        w_imag = 0.0
        theta = 2.0 * direction * math.pi / (2.0 * float(dual))
        s = math.sin(theta)
        t = math.sin(theta / 2.0)
        s2 = 2.0 * t * t
        for b in range(0, n, 2 * dual):
            i = 2 * b
            j = 2 * (b + dual)
            wd_real = data[j]
            wd_imag = data[j + 1]
            data[j] = data[i] - wd_real
            data[j + 1] = data[i + 1] - wd_imag
            data[i] += wd_real
            data[i + 1] += wd_imag
        for a in xrange(1, dual):
            tmp_real = w_real - s * w_imag - s2 * w_real
            tmp_imag = w_imag + s * w_real - s2 * w_imag
            w_real = tmp_real
            w_imag = tmp_imag
            for b in range(0, n, 2 * dual):
                i = 2 * (b + a)
                j = 2 * (b + a + dual)
                z1_real = data[j]
                z1_imag = data[j + 1]
                wd_real = w_real * z1_real - w_imag * z1_imag
                wd_imag = w_real * z1_imag + w_imag * z1_real
                data[j] = data[i] - wd_real
                data[j + 1] = data[i + 1] - wd_imag
                data[i] += wd_real
                data[i + 1] += wd_imag
        bit += 1
        dual *= 2


def FFT_bitreverse(N, data):
    n = N // 2
    nm1 = n - 1
    j = 0
    for i in range(nm1):
        ii = i << 1
        jj = j << 1
        k = n >> 1
        if i < j:
            tmp_real = data[ii]
            tmp_imag = data[ii + 1]
            data[ii] = data[jj]
            data[ii + 1] = data[jj + 1]
            data[jj] = tmp_real
            data[jj + 1] = tmp_imag
        while k <= j:
            j -= k
            k >>= 1
        j += k


def FFT_transform(N, data):
    FFT_transform_internal(N, data, -1)


def FFT_inverse(N, data):
    n = N / 2
    norm = 0.0
    FFT_transform_internal(N, data, +1)
    norm = 1 / float(n)
    for i in xrange(N):
        data[i] *= norm


def bench_FFT(loops, N, cycles):
    twoN = 2 * N
    init_vec = Random(7).RandomVector(twoN)
    range_it = xrange(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        x = copy_vector(init_vec)
        for i in xrange(cycles):
            FFT_transform(twoN, x)
            FFT_inverse(twoN, x)

    return pyperf.perf_counter() - t0


def add_cmdline_args(cmd, args):
    if args.benchmark:
        cmd.append(args.benchmark)


BENCHMARKS = {
    # function name => arguments
    'sor': (bench_SOR, 100, 10, Array2D),
    'sparse_mat_mult': (bench_SparseMatMult, 1000, 50 * 1000),
    'monte_carlo': (bench_MonteCarlo, 100 * 1000,),
    'lu': (bench_LU, 100,),
    'fft': (bench_FFT, 1024, 50),
}


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.argparser.add_argument("benchmark", nargs='?',
                                  choices=sorted(BENCHMARKS))

    args = runner.parse_args()
    if args.benchmark:
        benchmarks = (args.benchmark,)
    else:
        benchmarks = sorted(BENCHMARKS)

    for bench in benchmarks:
        name = 'scimark_%s' % bench
        print(name)
        args = BENCHMARKS[bench]
        (args[0])(10, *args[1:])
#        runner.bench_time_func(name, *args)



import sys
from pathlib import Path
from typing import Any, Dict
import covers as sc
import covers.branch as br
import ast
import atexit
import platform
import functools
import os
import tempfile
import json
import warnings

# Used for fork() support
input_tmpfiles = []
output_tmpfile = None


def fork_shim(sci):
    """Shims os.fork(), preparing the child to write its coverage to a temporary file
    and the parent to read from that file, so as to report the full coverage obtained.
    """
    original_fork = os.fork

    @functools.wraps(original_fork)
    def wrapper(*pargs, **kwargs):
        global input_tmpfiles, output_tmpfile

        tmp_file = tempfile.NamedTemporaryFile(
            mode="r+", encoding="utf-8", delete=False
        )

        if pid := original_fork(*pargs, **kwargs):
            input_tmpfiles.append(tmp_file)

        else:
            sci.signal_child_process()
            input_tmpfiles.clear()  # to be used by this process' children, if any
            output_tmpfile = tmp_file

        return pid

    return wrapper


def get_coverage(sci):
    """Combines this process' coverage with that of any previously forked children."""
    global input_tmpfiles, output_tmpfile

    cov = sci.get_coverage()
    if input_tmpfiles:
        for f in input_tmpfiles:
            try:
                fname = f.name
                f.seek(0, os.SEEK_END)
                # If the file is empty, it was likely closed, possibly upon exec
                if f.tell() != 0:
                    f.seek(0)
                    sc.merge_coverage(cov, json.load(f))
            except json.JSONDecodeError as e:
                warnings.warn(f"Error reading {fname}: {e}")
            finally:
                f.close()
                try:
                    os.remove(fname)
                except FileNotFoundError:
                    pass

    return cov


def exit_shim(sci):
    """Shims os._exit(), so a previously forked child process writes its coverage to
    a temporary file read by the parent.
    """
    original_exit = os._exit

    @functools.wraps(original_exit)
    def wrapper(*pargs, **kwargs):
        global output_tmpfile

        if output_tmpfile:
            json.dump(get_coverage(sci), output_tmpfile)
            output_tmpfile.flush()

        original_exit(*pargs, **kwargs)

    return wrapper


def merge_files(args, base_path):
    """Merges coverage files."""

    try:
        with args.merge[0].open() as jf:
            merged = json.load(jf)
    except Exception as e:
        warnings.warn(f"Error reading in {args.merge[0]}: {e}")
        return 1

    try:
        for f in args.merge[1:]:
            with f.open() as jf:
                sc.merge_coverage(merged, json.load(jf))
    except Exception as e:
        warnings.warn(f"Error merging in {f}: {e}")
        return 1

    try:
        with args.out.open("w", encoding="utf-8") as jf:
            if args.xml:
                sc.print_xml(
                    merged,
                    source_paths=[str(base_path)],
                    with_branches=args.branch,
                    xml_package_depth=args.xml_package_depth,
                    outfile=jf,
                )
            elif args.lcov:
                sc.print_lcov(
                    merged,
                    source_paths=[str(base_path)],
                    with_branches=args.branch,
                    outfile=jf,
                )
            else:
                json.dump(merged, jf, indent=(4 if args.pretty_print else None))

        # print human-readable table for merge results
        if not args.silent:
            sc.print_coverage(
                merged,
                outfile=sys.stdout,
                skip_covered=args.skip_covered,
                missing_width=args.missing_width,
            )

    except Exception as e:
        warnings.warn(str(e))
        return 1

    if args.fail_under:
        if merged["summary"]["percent_covered"] < args.fail_under:
            return 2

    return 0


def main():
    import argparse

    #
    # The intended usage is:
    #
    #   slipcover.py [options] (script | -m module [module_args...])
    #
    # but argparse doesn't seem to support this.  We work around that by only
    # showing it what we need.
    #
    ap = argparse.ArgumentParser(prog="Covers")
    ap.add_argument(
        "--branch", action="store_true", help="measure both branch and line coverage"
    )
    ap.add_argument("--json", action="store_true", help="select JSON output")
    ap.add_argument(
        "--pretty-print", action="store_true", help="pretty-print JSON output"
    )
    ap.add_argument("--xml", action="store_true", help="select XML output")
    ap.add_argument("--lcov", action="store_true", help="select LCOV output")
    ap.add_argument(
        "--xml-package-depth",
        type=int,
        default=99,
        help=(
            "Controls which directories are identified as packages in the report. "
            "Directories deeper than this depth are not reported as packages. "
            "The default is that all directories are reported as packages."
        ),
    )
    ap.add_argument("--out", type=Path, help="specify output file name")
    ap.add_argument("--source", help="specify directories to cover")
    ap.add_argument("--omit", help="specify file(s) to omit")
    ap.add_argument(
        "--immediate",
        action="store_true",
        help=(
            argparse.SUPPRESS
            if platform.python_implementation() == "PyPy"
            else "request immediate de-instrumentation"
        ),
    )
    ap.add_argument(
        "--skip-covered",
        action="store_true",
        help="omit fully covered files (from text, non-JSON output)",
    )
    ap.add_argument(
        "--fail-under",
        type=float,
        default=0,
        help="fail execution with RC 2 if the overall coverage lays lower than this",
    )
    ap.add_argument(
        "--threshold",
        type=int,
        default=50,
        metavar="T",
        help="threshold for de-instrumentation (if not immediate)",
    )
    ap.add_argument(
        "--missing-width",
        type=int,
        default=80,
        metavar="WIDTH",
        help="maximum width for `missing' column",
    )

    # intended for covers development only
    ap.add_argument("--silent", action="store_true", help=argparse.SUPPRESS)
    ap.add_argument("--dis", action="store_true", help=argparse.SUPPRESS)
    ap.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)
    ap.add_argument("--dont-wrap-pytest", action="store_true", help=argparse.SUPPRESS)
    ap.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s v{sc.__version__} (Python {'.'.join(map(str, sys.version_info[:3]))})",
    )

    g = ap.add_mutually_exclusive_group(required=True)
    g.add_argument("-m", dest="module", nargs=1, help="run given module as __main__")
    g.add_argument(
        "--merge",
        nargs="+",
        type=Path,
        help="merge JSON coverage files, saving to --out",
    )
    g.add_argument("script", nargs="?", type=Path, help="the script to run")
    ap.add_argument("script_or_module_args", nargs=argparse.REMAINDER)

    if "-m" in sys.argv:  # work around exclusive group not handled properly
        minus_m = sys.argv.index("-m")
        args = ap.parse_args(sys.argv[1 : minus_m + 2])
        args.script_or_module_args = sys.argv[minus_m + 2 :]
    else:
        args = ap.parse_args(sys.argv[1:])

    base_path = (
        Path(args.script).resolve().parent if args.script else Path(".").resolve()
    )

    if args.merge:
        if not args.out:
            ap.error("--out is required with --merge")
        return merge_files(args, base_path=base_path)

    file_matcher = sc.FileMatcher()

    if args.source:
        args.source = args.source.split(",")
        for s in args.source:
            file_matcher.addSource(s)
    elif args.script:
        file_matcher.addSource(Path(args.script).resolve().parent)

    if args.omit:
        for o in args.omit.split(","):
            file_matcher.addOmit(o)

    sci = sc.Covers(
        immediate=args.immediate,
        d_miss_threshold=args.threshold,
        branch=args.branch,
        disassemble=args.dis,
        source=args.source,
    )

    if not args.dont_wrap_pytest:
        sc.wrap_pytest(sci, file_matcher)

    if platform.system() != "Windows":
        os.fork = fork_shim(sci)
        os._exit = exit_shim(sci)

    def sci_atexit():
        global output_tmpfile

        def printit(coverage, outfile):
            if args.json:
                print(
                    json.dumps(coverage, indent=(4 if args.pretty_print else None)),
                    file=outfile,
                )
            elif args.xml:
                sc.print_xml(
                    coverage,
                    source_paths=[str(base_path)],
                    with_branches=args.branch,
                    xml_package_depth=args.xml_package_depth,
                    outfile=outfile,
                )
            elif args.lcov:
                sc.print_lcov(
                    coverage,
                    source_paths=[str(base_path)],
                    with_branches=args.branch,
                    outfile=outfile,
                )
            else:
                sc.print_coverage(
                    coverage,
                    outfile=outfile,
                    skip_covered=args.skip_covered,
                    missing_width=args.missing_width,
                )

        if not args.silent:
            coverage = get_coverage(sci)
            if args.out:
                with open(args.out, "w") as outfile:
                    printit(coverage, outfile)
            else:
                printit(coverage, sys.stdout)

    atexit.register(sci_atexit)

    if args.script:
        # python 'globals' for the script being executed
        script_globals: Dict[Any, Any] = dict()

        # needed so that the script being invoked behaves like the main one
        script_globals["__name__"] = "__main__"
        script_globals["__file__"] = args.script

        sys.argv = [str(args.script), *args.script_or_module_args]

        # the 1st item in sys.path is always the main script's directory
        sys.path.pop(0)
        sys.path.insert(0, str(base_path))

        with open(args.script, "r") as f:
            source = f.read()
            if args.branch and file_matcher.matches(args.script):
                t = br.preinstrument(source)
            else:
                t = ast.parse(source)
            code = compile(t, str(Path(args.script).resolve()), "exec")

        if file_matcher.matches(args.script):
            code = sci.instrument(code)

        with sc.ImportManager(sci, file_matcher):
            exec(code, script_globals)

    else:
        import runpy

        sys.argv = [*args.module, *args.script_or_module_args]
        with sc.ImportManager(sci, file_matcher):
            runpy.run_module(*args.module, run_name="__main__", alter_sys=True)

    if args.fail_under:
        cov = sci.get_coverage()
        if cov["summary"]["percent_covered"] < args.fail_under:
            return 2

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import ast
from typing import List, Union

# Import from Rust implementation
from .covers_core import encode_branch, analyze_branches_ts

EXIT = 0


def preinstrument(source: Union[str, ast.Module]) -> ast.Module:
    """Prepares an AST for coverage instrumentation using tree-sitter analysis.

    Args:
        source: The Python source code to analyze (string or AST Module)

    Returns:
        Modified AST with branch markers inserted
    """
    import textwrap

    # Handle both string and AST Module inputs
    if isinstance(source, ast.Module):
        # Unparse the AST to get source code for tree-sitter analysis
        source_str = ast.unparse(source)
        tree = source
    else:
        # Remove common leading whitespace from source string
        source_str = textwrap.dedent(source)
        tree = ast.parse(source_str)

    # Analyze source with tree-sitter (implemented in Rust)
    # Returns dict mapping branch_line -> [(insert_line, to_line), ...]
    # Note: This doesn't import ast in Rust, it's tree-sitter based!
    branch_data = analyze_branches_ts(source_str)

    class SlipcoverTransformer(ast.NodeTransformer):
        def __init__(self, branch_info):
            self.branch_info = branch_info
            self.source_lines = source_str.splitlines()

        def _mark_branch(self, from_line: int, to_line: int) -> List[ast.stmt]:
            # Using a constant Expr allows the compiler to optimize this to a NOP
            mark = ast.Expr(ast.Constant(None))
            for node in ast.walk(mark):
                node.lineno = node.end_lineno = encode_branch(from_line, to_line)  # type: ignore[attr-defined]
                # Leaving the columns unitialized can lead to invalid positions despite
                # our use of ast.fix_missing_locations
                node.col_offset = node.end_col_offset = -1  # type: ignore[attr-defined]

            return [mark]

        def _mark_branches_from_info(
            self, node: Union[ast.If, ast.For, ast.AsyncFor, ast.While], branch_markers
        ) -> ast.AST:
            # branch_markers is [(insert_line, to_line), ...]
            for idx, (insert_line, to_line) in enumerate(branch_markers):
                if insert_line == 0:
                    # This means append to orelse
                    node.orelse = self._mark_branch(node.lineno, to_line)
                elif idx == 0:
                    # First marker goes to body
                    node.body = self._mark_branch(node.lineno, insert_line) + node.body
                else:
                    # Subsequent markers go to orelse
                    if node.orelse:
                        node.orelse = (
                            self._mark_branch(node.lineno, insert_line) + node.orelse
                        )
                    else:
                        node.orelse = self._mark_branch(node.lineno, insert_line)

            super().generic_visit(node)
            return node

        def visit_If(self, node: ast.If) -> ast.AST:
            if node.lineno in self.branch_info:
                return self._mark_branches_from_info(
                    node, self.branch_info[node.lineno]
                )
            super().generic_visit(node)
            return node

        def visit_For(self, node: ast.For) -> ast.AST:
            if node.lineno in self.branch_info:
                return self._mark_branches_from_info(
                    node, self.branch_info[node.lineno]
                )
            super().generic_visit(node)
            return node

        def visit_AsyncFor(self, node: ast.AsyncFor) -> ast.AST:
            if node.lineno in self.branch_info:
                return self._mark_branches_from_info(
                    node, self.branch_info[node.lineno]
                )
            super().generic_visit(node)
            return node

        def visit_While(self, node: ast.While) -> ast.AST:
            if node.lineno in self.branch_info:
                return self._mark_branches_from_info(
                    node, self.branch_info[node.lineno]
                )
            super().generic_visit(node)
            return node

        def visit_Match(self, node: ast.Match) -> ast.Match:
            if node.lineno in self.branch_info:
                branch_markers = self.branch_info[node.lineno]

                # First markers go to each case body
                for idx, (insert_line, to_line) in enumerate(branch_markers):
                    if insert_line == 0:
                        # This is the wildcard case
                        node.cases.append(
                            ast.match_case(
                                ast.MatchAs(),
                                body=self._mark_branch(node.lineno, to_line),
                            )
                        )
                    elif idx < len(node.cases):
                        # Add marker to existing case
                        node.cases[idx].body = (
                            self._mark_branch(node.lineno, insert_line)
                            + node.cases[idx].body
                        )

            super().generic_visit(node)
            return node

    tree = SlipcoverTransformer(branch_data).visit(tree)
    ast.fix_missing_locations(tree)
    return tree


def preinstrument_and_compile(source: str, filename: str, branch: bool):
    """Parse, preinstrument, and compile Python source code.

    This function is called from Rust code to avoid importing ast in Rust.

    Args:
        source: Python source code
        filename: Filename for the code object
        branch: Whether to enable branch coverage

    Returns:
        Compiled code object
    """
    if branch:
        tree = preinstrument(source)
    else:
        tree = ast.parse(source, filename)

    return compile(tree, filename, "exec")
